'Module: SERVER
'File: server.bas
'*Purpose: Server and Connection objects which isolate the user from the ODBC API calls.
'*Dependencies: g_vb.bas, edit.bas, error.bas, list.bas, ascii.bas, api.bas, server.frm

'*NOTE: The programmer needs to only request a connection; login etc. is handled by
'*the objects.  Once a connection is requested it is referenced by a number.
'*After finishing with a connection it is "check in", and becomes available for reuse.
'To force a login, call cnInitializeServer.






Option Explicit

'*************************************************
' Connection Management Constants and Data Structures
'From the VB Demo

Type LoginType
    sServer As String                   ' Name of server
    sLogin As String                    ' Login string
    sPassword As String                 ' Password string
End Type

Type ServerPrivateType
    henv As Long                        ' Environment Variable
    nStatus As Integer                  ' Status of server connection
    sServerType As String               'type of server from ODBC driver data
    sLastUser As String                 'from ODBC.INI
    sDefaultDatabase As String          'database to initialize connection to
    lIsolationLevel As Long          'default isolation level
End Type

Type ServerType
    Login As LoginType
    PrivateData As ServerPrivateType
End Type

Type ConnectionType
    henv As Long                        ' Environment Variable
    sServer As String                   ' Server connected to
    hdbc As Long                        ' Connection handle
    hstmt As Long                       ' Statement handle
    nStatus As Integer                   ' Active or Inactive - communicating to server
    nUse As Integer                      ' Active or Inactive - Connection is being used by code or free to claim
    sTag As String                       ' A string of the form "DataSource <number>"
    lAutoCommit As Long                  'state of transaction auto-commit
    lIsolationLevel As Long              'current isolation level
    sActivityDescription As String       'String of what doing, for use by error handler
    nReturnCode As Integer              'Completion status of SQL activity
End Type

'Controls display of the login form

Type LoginStyleType
    nWithClerk As Integer
End Type

'Structure for passing information to/from the Login form.

Type PassLoginType
    Login As LoginType
    LoginStyle As LoginStyleType
End Type

'Type for parsing ALLBASE login string

Type AllBaseLoginType
    sSystem As String
    sServer As String
    sPath As String
    sLogin As String
    sPassword As String
End Type

'For retreival of column information

Type ColumnInfoType
    sColumnName As String
    lColumnLength As Long
End Type
                                       
'Declare the Connection

Dim connection() As ConnectionType   'list of requested connections.
Dim server() As ServerType           'list of requested servers.
Dim PassLogin As PassLoginType     'for passing info from login form.

'Holder for clerk initials
'- assume only one clerk per PC in the program at at time.
'- hence not needed as a part of the server structure

Dim sInitial As String               '

'constant for removing all server connections
Global Const sAll_SERVERS$ = "ALL"

'server types (in odbc.ini file)

Const ALLBASE$ = "HP ALLBASE/SQL"
Const SQL_SERVER$ = "SQL Server"

'Connection condition (status [relation to server] and use [by program])
 
 Const INACTIVE% = 0
 Const ACTIVE% = 1

'Allbase Locking level Flags

Global Const NO_LOCKS% = 0
Global Const REPEATABLE_READ_LOCK% = 1
Global Const READ_COMMIT_LOCK% = 2

'Maximum number of open connections

Const MAX_OPEN_CONNECTIONS% = 1
Const MAX_CONNECTIONS_ALLOWED% = 6

' ODBC Declarations and Constants

'SQLTransact values

Const AUTOCOMMIT_OFF& = 0
Const AUTOCOMMIT_ON& = 1

Global Const SQL_DATA_SOURCE_NAME = 6
Global Const SQL_USER_NAME = 8

'Connection Status

'Note, this value should not be the same as the
' ODBC errors in g_odbc
Global Const ODBC_ERROR_REPORTED% = 200

'Standard Declarations
'*********************************************************************/
' GLOBAL CONSTANTS '
'From SQL.H

' generally useful constants *
Global Const SQL_NTS% = -3                   ' NTS = Null Terminated String
Global Const SQL_SQLSTATE_SIZE% = 5          ' size of SQLSTATE
Global Const SQL_MAX_MESSAGE_LENGTH% = 512   ' message buffer size
Global Const SQL_MAX_DSN_LENGTH% = 32        ' maximum data source name size

'Return Coded from the Functions
' RETCODEs *
Global Const SQL_ERROR% = -1
Global Const SQL_INVALID_HANDLE% = -2
Global Const SQL_NEED_DATA% = 99
Global Const SQL_NO_DATA_FOUND% = 100
Global Const SQL_SUCCESS% = 0
Global Const SQL_SUCCESS_WITH_INFO% = 1

' SQLFreeStmt defines *
Global Const SQL_CLOSE% = 0
Global Const SQL_DROP% = 1
Global Const SQL_UNBIND% = 2
Global Const SQL_RESET_PARAMS% = 3

' SQLSetParam defines *
Global Const SQL_C_DEFAULT% = 99

' SQLTransact defines *
Global Const SQL_COMMIT% = 0
Global Const SQL_ROLLBACK% = 1

' Standard SQL datatypes, using ANSI type numbering *
Global Const SQL_CHAR% = 1
Global Const SQL_NUMERIC% = 2
Global Const SQL_DECIMAL% = 3
Global Const SQL_INTEGER% = 4
Global Const SQL_SMALLINT% = 5
Global Const SQL_FLOAT% = 6
Global Const SQL_REAL% = 7
Global Const SQL_DOUBLE% = 8
Global Const SQL_VARCHAR% = 12

'#define SQL_C_CHAR    SQL_CHAR       /* CHAR, VARCHAR, DECIMAL, NUMERIC */
'#define SQL_C_LONG    SQL_INTEGER    /* INTEGER          */
'#define SQL_C_SHORT   SQL_SMALLINT   /* SMALLINT         */
'#define SQL_C_FLOAT   SQL_REAL       /* REAL             */
'#define SQL_C_DOUBLE  SQL_DOUBLE     /* FLOAT, DOUBLE    */

Global Const SQL_C_CHAR% = 1         '/* CHAR, VARCHAR, DECIMAL, NUMERIC */
Global Const SQL_C_LONG% = 4         ' SQL_INTEGER    /* INTEGER          */
Global Const SQL_C_SHORT% = 5        ' SQL_SMALLINT   /* SMALLINT         */
Global Const SQL_C_FLOAT% = 7        ' SQL_REAL       /* REAL             */
Global Const SQL_C_DOUBLE% = 8       ' SQL_DOUBLE     /* FLOAT, DOUBLE    */

Global Const SQL_TYPE_MIN% = 1
Global Const SQL_TYPE_NULL% = 0
Global Const SQL_TYPE_MAX% = 12


' * NULL status constants.  These are used in SQLColumns, SQLColAttributes,
' SQLDescribeCol, and SQLSpecialColumns to describe the nullablity of a
' column in a table.  SQL_NULLABLE_UNKNOWN can be returned only by
' SQLDescribeCol or SQLColAttributes.  It is used when the DBMS's meta-data
' does not contain this info.  *
Global Const SQL_NO_NULLS% = 0
Global Const SQL_NULLABLE% = 1
Global Const SQL_NULLABLE_UNKNOWN% = 2

' Special length values *
Global Const SQL_NULL_DATA% = -1
Global Const SQL_DATA_AT_EXEC% = -2

' SQLColAttributes defines *
Global Const SQL_COLUMN_COUNT% = 0
Global Const SQL_COLUMN_NAME% = 1
Global Const SQL_COLUMN_TYPE% = 2
Global Const SQL_COLUMN_LENGTH% = 3
Global Const SQL_COLUMN_PRECISION% = 4
Global Const SQL_COLUMN_SCALE% = 5
Global Const SQL_COLUMN_DISPLAY_SIZE% = 6
Global Const SQL_COLUMN_NULLABLE% = 7
Global Const SQL_COLUMN_UNSIGNED% = 8
Global Const SQL_COLUMN_MONEY% = 9
Global Const SQL_COLUMN_UPDATABLE% = 10
Global Const SQL_COLUMN_AUTO_INCREMENT% = 11
Global Const SQL_COLUMN_CASE_SENSITIVE% = 12
Global Const SQL_COLUMN_SEARCHABLE% = 13
Global Const SQL_COLUMN_TYPE_NAME% = 14

' SQLColAttributes subdefines for SQL_COLUMN_UPDATABLE *
Global Const SQL_ATTR_READONLY% = 0
Global Const SQL_ATTR_WRITE% = 1
Global Const SQL_ATTR_READWRITE_UNKNOWN% = 2

' SQLColAttributes subdefines for SQL_COLUMN_SEARCHABLE *
' These are also used by SQLGetInfo                     *
Global Const SQL_UNSEARCHABLE% = 0
Global Const SQL_LIKE_ONLY% = 1
Global Const SQL_ALL_EXCEPT_LIKE% = 2
Global Const SQL_SEARCHABLE% = 3

' SQLError defines *
Global Const SQL_NULL_HENV% = 0
Global Const SQL_NULL_HDBC% = 0
Global Const SQL_NULL_HSTMT% = 0

' options for SQLSetConnectOption/SQLGetConnectOption
'
Global Const SQL_ACCESS_MODE = 101
Global Const SQL_AUTOCOMMIT = 102
Global Const SQL_LOGIN_TIMEOUT = 103
Global Const SQL_OPT_TRACE = 104
Global Const SQL_OPT_TRACEFILE = 105
Global Const SQL_TRANSLATE_DLL = 106
Global Const SQL_TRANSLATE_OPTION = 107
Global Const SQL_TXN_ISOLATION = 108
Global Const SQL_CONNECT_OPT_DRVR_START = 1000

'Locking Levels

Global Const SQL_TXN_READ_UNCOMMITTED& = &H1&
Global Const SQL_TXN_READ_COMMITTED& = &H2&
Global Const SQL_TXN_REPEATABLE_READ& = &H4&
Global Const SQL_TXN_SERIALIZABLE& = &H8&
Global Const SQL_TXN_VERSIONING& = &H10&

'SQLGetInfo

Global Const SQL_DEFAULT_TXN_ISOLATION = 26

'***************************************************************
'Function Prototype Declarations
'Commented out functions are currently not used

'CORE FUNCTIONS

Declare Function SQLAllocConnect Lib "odbc.dll" (ByVal env As Long, hdbc As Long) As Integer

Declare Function SQLAllocEnv Lib "odbc.dll" (env As Long) As Integer

Declare Function SQLAllocStmt Lib "odbc.dll" (ByVal hdbc As Long, hstmt As Long) As Integer

'Declare Function SQLBindCol

'Declare Function SQLCancel

'Declare Function SQLColAttributes

Declare Function SQLConnect Lib "odbc.dll" (ByVal hdbc As Long, ByVal server As String, ByVal serverlen As Integer, ByVal uid As String, ByVal uidlen As Integer, ByVal pwd As String, ByVal pwdlen As Integer) As Integer

Declare Function SQLDescribeCol Lib "odbc.dll" (ByVal hstmt As Long, ByVal colnum As Integer, ByVal colname As String, ByVal Buflen As Integer, colnamelen As Integer, dtype As Integer, dl As Long, ds As Integer, n As Integer) As Integer

Declare Function SQLDisconnect Lib "odbc.dll" (ByVal hdbc As Long) As Integer

Declare Function SQLError Lib "odbc.dll" (ByVal env As Long, ByVal hdbc As Long, ByVal hstmt As Long, ByVal SQLState As String, NativeError As Long, ByVal Buffer As String, ByVal Buflen As Integer, Outlen As Integer) As Integer

Declare Function SQLExecDirect Lib "odbc.dll" (ByVal hstmt As Long, ByVal sqlString As String, ByVal sqlstrlen As Long) As Integer

'Declare Function SQLExecute

Declare Function SQLFetch Lib "odbc.dll" (ByVal hstmt As Long) As Integer

Declare Function SQLFreeConnect Lib "odbc.dll" (ByVal hdbc As Long) As Integer

Declare Function SQLFreeEnv Lib "odbc.dll" (ByVal env As Long) As Integer

Declare Function SQLFreeStmt Lib "odbc.dll" (ByVal hstmt As Long, ByVal EndOption As Integer) As Integer

'Declare Function SQLGetCursorName

Declare Function SQLNumREsultCols Lib "odbc.dll" (ByVal hstmt As Long, NumCols As Integer) As Integer

'Declare Function SQLPrepare

'Declare Function SQLRowCount

'Declare Function SQLSetCursorName

'Declare Function SQLSetParam

Declare Function SQLTransact Lib "odbc.dll" (ByVal henv As Long, ByVal hdbc As Long, ByVal fType As Integer) As Integer

'********************************************************
'LEVEL 1 FUNCTIONS

'Declare Function SQLColumns

Declare Function SQLDriverConnect Lib "odbc.dll" (ByVal hdbc As Long, ByVal hwnd As Integer, ByVal szCSIn As String, ByVal cbCSIn As Integer, ByVal szCSOut As String, ByVal cbCSMax As Integer, cbCSOut As Integer, ByVal f As Integer) As Integer

'Declare Function SQLGetConnectOption

Declare Function SQLGetData Lib "odbc.dll" (ByVal hstmt As Long, ByVal col As Integer, ByVal wConvType As Integer, ByVal lpbBuf As String, ByVal dwbuflen As Long, lpcbout As Long) As Integer

'Declare Function SQLGetFunctions

Declare Function SQLGetInfo Lib "odbc.dll" (ByVal hdbc As Long, ByVal hwnd As Integer, ByVal szInfo As String, ByVal cbInfoMax As Integer, cbInfoOut As Integer) As Integer

'Declare Function SQLGetStmtOption

'Declare Function SQLGetTypeInfo

'Declare Function SQLParamData

'Declare Function SQLPutData

Declare Function SQLSetConnectOption Lib "odbc.dll" (ByVal hdbc As Long, ByVal fOption As Integer, ByVal vParam As Long) As Integer

'Declare Function SQLSetStmtOption

'Declare Function SQLSpecialColumns

'Declare Function SQLStatistics

Declare Function SQLTables Lib "odbc.dll" (ByVal hstmt As Long, ByVal q As Long, ByVal cbq As Integer, ByVal o As Long, ByVal cbo As Integer, ByVal t As Long, ByVal cbt As Integer, ByVal tt As Long, ByVal cbtt As Integer) As Integer

'Level 2

Declare Function SQLMoreResults Lib "odbc.dll" (ByVal hstmt As Long) As Integer

Declare Function SQLRowCount Lib "odbc.dll" (ByVal hstmt As Long, pcrow As Long) As Integer

Function clRowsEffected (nConnectionNumber As Integer, sCmd As String, nReturnStatus As Integer) As Long

'*Purpose: Return the count of rows affected by an SQL execution (delete, update).

'Declarations

    Dim nStatus As Integer
    Dim nStatus2 As Integer

    Dim lRows As Long

'End Declarations

    nReturnStatus% = FAIL%

'Process SQL

    nStatus% = cnNoReturn(nConnectionNumber%, sCmd$)
    If nStatus% = FAIL Then Exit Function

'Get count of rows effected
    connection(nConnectionNumber%).nReturnCode = SQLRowCount(connection(nConnectionNumber%).hstmt, lRows&)
 ' SQLRowCount(Connection.hstmt, pcrow&)
    
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
        nStatus% = cnODBCError(nConnectionNumber%)
        nStatus2% = cnErrorHandle(SQL_SYNTAX_ERROR%, SERVER_OBJECT_GROUP%, "cnRowsEffected", sCmd$ & NL$ & "Contact Programmer")

        Select Case nStatus
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select

    End If
    
    clRowsEffected& = lRows&
    
    nReturnStatus% = SUCCEED%
    
End Function

Private Function cnActivateConnection (nConnectionNumber As Integer) As Integer

'*Purpose: Activate a specified connection (connect to server)

'Declarations

    Dim HoldConnection As ConnectionType
    Dim AllBaseLogin As AllBaseLoginType

    Dim nStatus As Integer
    Dim nIndex As Integer

    Dim lDataBase As Long

    Dim sServer As String
    Dim sLogin As String
    Dim sPassword As String

'End Declarations
    
    nStatus% = cnGetServerIndex(connection(nConnectionNumber%).sServer$, nIndex%)
    If nStatus% = FAIL% Then
        Exit Function
    End If

' initialize other values

    connection(nConnectionNumber%).hdbc = SQL_NULL_HDBC
    connection(nConnectionNumber%).hstmt = SQL_NULL_HSTMT
  
'  Allocate a connection handle

    connection(nConnectionNumber%).nReturnCode = SQLAllocConnect(connection(nConnectionNumber%).henv, connection(nConnectionNumber%).hdbc)
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
'
'Test connection failure
'   Save needed values to a hold connection.
        
        HoldConnection = connection(nConnectionNumber%)

        nStatus% = cnODBCError(nConnectionNumber%)

'Save return value and restore other data
        
        HoldConnection.nReturnCode = connection(nConnectionNumber%).nReturnCode
        connection(nConnectionNumber%) = HoldConnection
      
        Select Case nStatus
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select
    End If

  
'  Make the connection

    Select Case UCase(server(nIndex%).PrivateData.sServerType$)
    Case UCase(SQL_SERVER$)
        lDataBase& = connection(nConnectionNumber%).hdbc
        sServer$ = server(nIndex%).Login.sServer$
        sLogin$ = server(nIndex%).Login.sLogin$
        sPassword$ = server(nIndex%).Login.sPassword$

    Case UCase(ALLBASE$)
        lDataBase& = connection(nConnectionNumber%).hdbc
        sServer$ = server(nIndex%).Login.sServer$
        sPassword$ = server(nIndex%).Login.sPassword$
        sLogin$ = server(nIndex%).PrivateData.sLastUser$
    End Select

    
    
    connection(nConnectionNumber%).nReturnCode = SQLConnect(lDataBase&, sServer$, Len(sServer$), sLogin$, Len(sLogin$), sPassword$, Len(sPassword$))
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
        HoldConnection = connection(nConnectionNumber%)

        nStatus% = cnODBCError(nConnectionNumber%)
        Select Case nStatus%
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select

'Test connection success
'Save return value and restore other data
        
        HoldConnection.nReturnCode = connection(nConnectionNumber%).nReturnCode
        
        connection(nConnectionNumber%) = HoldConnection
      
    End If

  
'  Allocate a statement handle
    
    connection(nConnectionNumber%).nReturnCode = SQLAllocStmt(connection(nConnectionNumber%).hdbc, connection(nConnectionNumber%).hstmt)
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
        nStatus% = cnODBCError(nConnectionNumber%)
      
        Select Case nStatus
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select
    End If

'Initialize of Connection

    Select Case server(nIndex).PrivateData.sServerType$
    Case SQL_SERVER$
'Initialize default database
        nStatus% = cnSetDatabase(server(nIndex%).PrivateData.sDefaultDatabase$, nConnectionNumber%)
        If nStatus% = FAIL% Then
            nStatus% = cnErrorHandle(NO_DEFAULT_DATABASE%, SERVER_OBJECT_GROUP%, "cnCreateConnection", "Failure Initializing Connection to Default Database" & NL$ & "Contact Programmer")
        End If
    End Select


'Set Flag - active relationship to server
    
    connection(nConnectionNumber%).lAutoCommit& = AUTOCOMMIT_ON&
    
    connection(nConnectionNumber%).nStatus% = ACTIVE%
    
    cnActivateConnection% = SUCCEED%

End Function

Function cnAllbaseConnection (nConnection As Integer, sActivityDescription As String) As Integer

'*Purpose: Start a connection in AllBase Default mode (Read Commit, controlled transaction)

'Declarations

    Dim nStatus As Integer

'End Declarations

    nStatus% = cnOpenConnection(nConnection%, sActivityDescription$)
    If nStatus% = FAIL% Then
        Exit Function
    End If

    nStatus% = cnBeginTrans(nConnection%, READ_COMMIT_LOCK%)
    If nStatus% = FAIL% Then
        Exit Function
    End If

    cnAllbaseConnection% = SUCCEED%

End Function

Private Function cnAllocateEnvironment (ThisServer As ServerType) As Integer

'*Purpose: Allocate environmental variable space.  Internal to server object creation.

'Declarations

    Dim nStatus As Integer

'End Declarations
  
'  Allocate the memory for environment handler and initializes ODBC call-level interface.
'
    nStatus% = SQLAllocEnv(ThisServer.PrivateData.henv)
    If nStatus% <> SQL_SUCCESS% Then
        nStatus% = cnErrorHandle(MEMORY_ALLOCATION_ERROR%, SERVER_OBJECT_GROUP%, "cnAllocateEnvironment", "Memory Allocation Error.")
        Select Case nStatus%
        Case EXIT_STATUS%
            Exit Function
        End Select
    End If
    
    cnAllocateEnvironment% = SUCCEED%

End Function

Function cnBeginTrans (nConnectionNumber As Integer, nLockFlag As Integer) As Integer

'*Purpose: Begin a sql transaction

'Declarations

    Dim nStatus As Integer
    Dim nIndex As Integer

    Dim lLock As Long

    Dim sCmd As String

'End Declarations

    nStatus% = cnGetServerIndex(connection(nConnectionNumber%).sServer$, nIndex%)
    If nStatus% = FAIL% Then
        Exit Function
    End If

    connection(nConnectionNumber%).nReturnCode = SQLSetConnectOption(connection(nConnectionNumber%).hdbc, SQL_AUTOCOMMIT, AUTOCOMMIT_OFF&)
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
    nStatus% = cnODBCError(nConnectionNumber%)
        Select Case nStatus%
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select
    End If
    
    connection(nConnectionNumber%).lAutoCommit& = AUTOCOMMIT_OFF&

'    Select Case UCase(server(nIndex).PrivateData.sServerType$)
'    Case UCase(SQL_SERVER$)
'        sCmd$ = "begin tran"
'        nStatus% = cnNoReturn(nConnectionNumber%, sCmd$)
'        If nStatus% = FAIL% Then
'            Exit Function
'        End If

'    Case UCase(ALLBASE$)
'        sCmd$ = "begin work"
'Add locking parameter
        Select Case nLockFlag%
        Case NO_LOCKS%
        Case REPEATABLE_READ_LOCK%
            lLock& = SQL_TXN_REPEATABLE_READ
        Case READ_COMMIT_LOCK%
            lLock& = SQL_TXN_READ_COMMITTED
        End Select
        
'Specified locking
        If lLock& > 0 Then
            connection(nConnectionNumber%).nReturnCode = SQLSetConnectOption(connection(nConnectionNumber%).hdbc, SQL_TXN_ISOLATION, lLock&)
            If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
                nStatus% = cnODBCError(nConnectionNumber%)
                Select Case nStatus%
                Case EXIT_STATUS%, FAIL%
                    Exit Function
                End Select
            End If

            connection(nConnectionNumber%).lIsolationLevel& = lLock&
        End If

'    End Select

    
    cnBeginTrans% = SUCCEED%

End Function

Private Function cnCheckOpenServer (sThisServer As String) As Integer

'*Purpose: Check array of servers to see if already connected

'Declarations

    Dim x As Integer
    Dim nStatus As Integer

    Dim nLast As Integer

'End Declarations

'If no servers, then nothing to check
    
    On Error GoTo nCheckOpenServer_Error

    nLast% = UBound(server)

    On Error GoTo 0

    For x% = 0 To nLast%
        If server(x%).Login.sServer$ = sThisServer$ Then
            If server(x%).PrivateData.nStatus% = ACTIVE% Then
                cnCheckOpenServer% = SUCCEED%
                Exit Function
            End If
        End If
    Next x%

nCheckOpenServer_Exit:

    Exit Function

nCheckOpenServer_Error:

    Resume nCheckOpenServer_Exit

End Function

Function cnCheckStatus (nConnectionNumber As Integer, nCode As Integer) As Integer

'*Purpose: Check return status of SQL activity, and determine action.

'Declarations

    Dim nStatus As Integer

'End Declarations

    If connection(nConnectionNumber%).nReturnCode% <> nCode% Then
        nStatus% = cnODBCError(nConnectionNumber%)
        Select Case nStatus%
        Case EXIT_STATUS%
            Exit Function
        End Select
    End If

    cnCheckStatus% = SUCCEED%

End Function

Function cnCommitTrans (nConnectionNumber As Integer) As Integer

'*Purpose: Commit a SQL transaction

'Declarations

    Dim nStatus As Integer
    Dim nIndex As Integer

    Dim sCmd As String

'End Declarations
 
    nStatus% = cnGetServerIndex(connection(nConnectionNumber%).sServer$, nIndex%)
    If nStatus% = FAIL% Then
        Exit Function
    End If

'    Select Case UCase(server(nIndex%).PrivateData.sServerType$)
'    Case UCase(SQL_SERVER$)
'        sCmd$ = "commit tran"
'        nStatus% = cnNoReturn(nConnectionNumber%, sCmd$)
'    Case UCase(ALLBASE$)
        connection(nConnectionNumber%).nReturnCode% = SQLTransact(connection(nConnectionNumber%).henv, connection(nConnectionNumber%).hdbc, SQL_COMMIT)
        If connection(nConnectionNumber%).nReturnCode% <> SQL_SUCCESS% Then
            nStatus% = cnODBCError(nConnectionNumber%)
        End If
'    End Select
    
    Select Case nStatus%
    Case EXIT_STATUS%, FAIL%
        Exit Function
    End Select
    
    cnCommitTrans% = SUCCEED%
    
End Function

Private Function cnCreateNewConnection (sThisServer As String, nThisConnection As Integer) As Integer

'*Purpose: Create a new connection in the server-object for the server (do not connect to server).

'Declarations

    Dim nMaxConnections As Integer
    Dim nIndex As Integer

    Dim sServer As String
    Dim sLogin As String
    Dim sPassword As String

    Dim x As Integer
    Dim nStatus As Integer

'End Declarations

'Get server index
'Can assume it exists, as was checked in nOpenConnectionMulti, which calls this function

    nStatus% = cnGetServerIndex(sThisServer$, nIndex%)
    If nStatus% = FAIL% Then
        Exit Function
    End If

'Re Dimension connection array for new connection


    On Error GoTo nCreateNewConnection_Error

    nMaxConnections% = UBound(connection) + 1

    On Error GoTo 0

'Check not opening connections and not returning
'Not exceed max allowed

    If nMaxConnections% >= MAX_CONNECTIONS_ALLOWED% Then
        nStatus% = cnErrorHandle(MAX_CONNECTIONS_EXCEEDED%, SERVER_OBJECT_GROUP%, "cnCreateNewConnection", "Program Exceeds Allowed Connections (" & Format$(nMaxConnections%, "0") & " Contact Programmer")
        nMaxConnections% = nMaxConnections% - 1
        Exit Function
    End If
        

    ReDim Preserve connection(nMaxConnections%) As ConnectionType

    connection(nMaxConnections%).sServer$ = server(nIndex%).Login.sServer$
    connection(nMaxConnections%).henv = server(nIndex%).PrivateData.henv
    
    nThisConnection% = nMaxConnections%

    cnCreateNewConnection% = SUCCEED%

    Exit Function

nCreateNewConnection_Error:

    nMaxConnections% = 0

    Resume Next

End Function

Private Function cnDeAllocateEnvironment (sThisServer As String) As Integer

'*Purpose: Allocate environmental variable space

'Declarations

    Dim x As Integer
    Dim y As Integer
    Dim nStatus As Integer
    Dim nRemoveFlag As Integer
    Dim nLast As Integer
    

'End Declarations

    On Error GoTo nDeAllocateEnvironment_Error

    nLast% = UBound(server)

    On Error GoTo 0

    For x% = 0 To nLast
        
        nRemoveFlag% = False

        Select Case server(x%).Login.sServer
        Case Is = sThisServer$
            nRemoveFlag% = True
        Case Is = sAll_SERVERS$
            nRemoveFlag% = True
        End Select

        If nRemoveFlag% Then

            nStatus% = SQLFreeEnv(server(x%).PrivateData.henv)
            If nStatus% <> SQL_SUCCESS% Then
                For y% = 0 To UBound(connection)
'Shut down server connections
                    If connection(y%).sServer$ = server(x%).Login.sServer$ Then
                        connection(y%).hdbc = SQL_NULL_HDBC
                        connection(y%).hstmt = SQL_NULL_HSTMT
                        connection(y%).nStatus% = INACTIVE%
                    End If
                Next y%
            End If

            nStatus% = cnODBCError(x%)

            Select Case nStatus
            Case EXIT_STATUS%, FAIL%
                Exit Function
            End Select
            
            server(x%).PrivateData.henv = SQL_NULL_HENV

            server(x%).PrivateData.nStatus% = INACTIVE%

        End If

    Next x%

    cnDeAllocateEnvironment% = SUCCEED%

nDeAllocateEnvironment_Exit:

    Exit Function

nDeAllocateEnvironment_Error:

    Resume nDeAllocateEnvironment_Exit

End Function

Private Function cnFindInactiveConnection (sThisServer As String, nThisConnection As Integer) As Integer

'*Purpose: Search connection list for inactive (not checked out) connection on specified server.

'Declarations

    Dim x As Integer

    Dim nLast As Integer

'End Declarations
    
'Trap no allocated connections
    
    On Error GoTo nCheckOpenConnection_Error

    nLast% = UBound(connection)

    On Error GoTo 0
    
    For x% = 0 To nLast%
        If connection(x%).sServer = sThisServer$ Then
            Select Case connection(x%).nUse%
            Case INACTIVE%
                nThisConnection% = x%
                cnFindInactiveConnection% = SUCCEED%
                Exit Function
            End Select
        End If
    Next x%

nCheckOpenConnection_Exit:
    
    Exit Function

nCheckOpenConnection_Error:

    Resume nCheckOpenConnection_Exit
                       
End Function

Function cnGetColumnInfo (nConnectionNumber As Integer, ColumnInfo() As ColumnInfoType) As Integer

'*Purpose: Get info about result set columns and put into info array (one record per column).

'Declarations

    Dim nStatus As Integer
    Dim nColumns As Integer
    Dim x As Integer

    Dim a As Integer
    Dim b As Integer
    Dim d As Integer
    Dim e As Integer

'End Declarations

    ReDim ColumnInfo(0) As ColumnInfoType

'Get number of columns
    
    nStatus% = SQLNumREsultCols(connection(nConnectionNumber%).hstmt, nColumns%)
    
    If nStatus% <> SQL_SUCCESS% Then
        nColumns% = 0
        Exit Function
    End If

    For x% = 0 To nColumns% - 1

        ReDim Preserve ColumnInfo(x%) As ColumnInfoType

        ColumnInfo(x%).sColumnName$ = Space$(30)

        nStatus% = SQLDescribeCol(connection(nConnectionNumber%).hstmt, x% + 1, ColumnInfo(x%).sColumnName$, Len(ColumnInfo(x%).sColumnName$), a%, b%, ColumnInfo(x%).lColumnLength&, d%, e%)
        If nStatus% <> SQL_SUCCESS% Then
            ColumnInfo(x%).sColumnName$ = BLANK$
            Exit Function
        End If

    Next x%

    cnGetColumnInfo% = SUCCEED%

End Function

Function cnGetDefaultIsolation (nConnection As Integer) As Integer

'*Purpose Read the default isolation level for the server
'Modified 1/24/93 hmw 1) fix sHold by initializing to a size
'   2) SQLGetInfo's return is a pointer to a string, 16 or 32 bit number, or a 32 bit map
'        locking is a 32 bit mask - treat as a long via Asc(sHold)
'Modified 1/27/94 hmw jsp
'  Per memo from Teresa Huxford, the Allbase default isolation level is hard coded per the following
'Declarations

    Dim nStatus As Integer
    Dim nIndex As Integer

    Dim nSize As Integer

    Dim sHold As String

'End Declarations

    sHold$ = Space$(256)

    nStatus% = cnGetServerIndex(connection(nConnection%).sServer$, nIndex%)
    If nStatus% = FAIL% Then
        Exit Function
    End If

    Select Case UCase(server(nIndex).PrivateData.sServerType$)
    Case UCase(ALLBASE$)
        server(nIndex%).PrivateData.lIsolationLevel& = SQL_TXN_READ_COMMITTED&
    Case UCase(SQL_SERVER$)
        connection(nConnection%).nReturnCode% = SQLGetInfo(connection(nConnection%).hdbc, SQL_DEFAULT_TXN_ISOLATION, sHold$, Len(sHold$), nSize%)
        If connection(nConnection%).nReturnCode% <> SQL_SUCCESS% Then
        nStatus% = cnODBCError(nConnection%)
            Select Case nStatus%
            Case EXIT_STATUS%, FAIL%
                Exit Function
            End Select
        End If

        server(nIndex%).PrivateData.lIsolationLevel& = Asc(sHold$)
    End Select

    cnGetDefaultIsolation% = SUCCEED%

End Function

Function cnGetInitials (sTmpInitial As String) As Integer

'*Purpose: Retreive initials from the structure

    sTmpInitial$ = sInitial$

    cnGetInitials% = SUCCEED%

End Function

Function cnGetLoginData (ThisLogin As LoginType) As Integer

'*Purpose: Get information about a server login

'Declarations

    Dim x As Integer

    Dim nLast As Integer

'End Declarations

'Trap error if no servers allocated in array

    On Error GoTo nGetLoginData_Error

    nLast = UBound(server)

    On Error GoTo 0

    For x% = 0 To nLast

        If ThisLogin.sServer$ = server(x%).Login.sServer$ Then
            ThisLogin = server(x%).Login
            cnGetLoginData% = SUCCEED%
            Exit Function
        End If
    Next x%

nGetLoginData_Exit:

    Exit Function

nGetLoginData_Error:

    Resume nGetLoginData_Exit

End Function

Function cnGetPassLogin (ThisServer As PassLoginType) As Integer

'*Purpose: Used by login form to get login information and form style info.

    ThisServer = PassLogin

    cnGetPassLogin% = SUCCEED%

End Function

Function cnGetResults (nConnectionNumber As Integer) As Integer

'*Purpose: Check if more result sets are pending

'Declarations

    Dim nStatus As Integer

'End Declarations

    nStatus% = SQLMoreResults(connection(nConnectionNumber).hstmt)

    cnGetResults% = nStatus

End Function

Function cnGetRow (nThisConnection As Integer) As Integer

'*Purpose: imbed ODBC next row in a library

'Declarations

    Dim nStatus As Integer
    Dim nTmpStatus As Integer

'End Declarations

    nTmpStatus% = SQLFetch%(connection(nThisConnection%).hstmt)

    
    If nTmpStatus% <> SQL_SUCCESS% Then
        Select Case UCase(connection(nThisConnection%).sServer$)
        Case UCase(ALLBASE$)

'When finish retreival, ALLBASE needs a commit to free the read locks.
'NOTE: hmw 9/16/93 This is due to a bug in the ALLBASE Beta ODBC driver, and should not be needed in future releases

            nStatus% = cnCommitTrans(nThisConnection%)
        End Select
    End If

    connection(nThisConnection%).nReturnCode = nTmpStatus%
    cnGetRow% = nTmpStatus%

End Function

Function cnGetRowCount (nConnectionNumber As Integer, lRowCount As Long) As Integer

'*Purpose: Get number of rows effected by SQL call; NOTE: This API call only works for update/insert or delete.

'Declarations

    Dim nStatus As Integer

'End Deaclarations

    lRowCount& = 0

    connection(nConnectionNumber%).nReturnCode = SQLRowCount(connection(nConnectionNumber%).hstmt, lRowCount&)
    
    nStatus% = cnCheckStatus%(nConnectionNumber%, SQL_SUCCESS%)
    If nStatus% = FAIL% Then
        nStatus% = cnODBCError(nConnectionNumber%)
        Select Case nStatus
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select

    End If
    
End Function

Function cnGetServerIndex (sServer As String, nIndex As Integer) As Integer

'*Purpose: Get information about a server

'Declarations

    Dim nLast As Integer

'End Declarations

'Trap error if no servers allocated in array

    On Error GoTo nGetServerIndex_Error

    nLast = UBound(server)

    On Error GoTo 0

    
    
    For nIndex% = 0 To nLast

        If sServer$ = server(nIndex%).Login.sServer$ Then
            If server(nIndex%).PrivateData.nStatus% = ACTIVE% Then
                cnGetServerIndex% = SUCCEED%
                Exit Function
            End If
        End If
    Next nIndex%

    nIndex% = -1

nGetServerIndex_Exit:

    Exit Function

nGetServerIndex_Error:

    Resume nGetServerIndex_Exit

End Function

Function cnGetServerType (sServer As String, sServerType As String) As Integer

'*Purpose: Read server type from ODBC.INI file.

'Declarations

    Const FileName = "ODBC.INI"

    Dim nStatus As Integer

    Dim sTemp As String * 255

'End Declarations

'Get Server type

    nStatus% = GetPrivateProfileString("ODBC Data Sources", sServer$, "n/a", sTemp$, Len(sTemp$), FileName$)

    sServerType = csRStrip$(sTemp$)

    cnGetServerType% = SUCCEED%

End Function

Function cnInitializeServer (TmpLogin As LoginType, sThisDatabase As String, ThisStyle As LoginStyleType) As Integer

'*Purpose: Initialize server by creating a connection then releasing it.

'Declarations
    
    Dim x As Integer
    Dim nStatus As Integer
    Dim nIndex As Integer
    Dim nConnection As Integer

'End Declarations

'Check for server existing
    
    nStatus% = cnOpenServer(TmpLogin)
    If nStatus% = FAIL% Then
        Exit Function
    End If

'Set Default database
'Should be open, so get index

    nStatus% = cnGetServerIndex(TmpLogin.sServer$, nIndex%)
    If nStatus% = FAIL% Then
        Exit Function
    End If

    server(nIndex%).PrivateData.sDefaultDatabase$ = sThisDatabase$

'Set variable for displaying clerk initital on login form

    PassLogin.LoginStyle = ThisStyle

'Get and return connection to force login display.
'Test connection login
    
    nStatus% = cnOpenConnectionMulti(TmpLogin.sServer$, nConnection%, "Initialize Server")
    
    If nStatus% = SUCCEED% Then
        nStatus% = cnGetDefaultIsolation(nConnection%)
        nStatus% = cnReturnConnection(nConnection%)
    Else
'BOS Applications, don't give retry
        If server(nIndex%).PrivateData.sServerType$ = ALLBASE$ Then
            Exit Function
        End If

        TmpLogin.sPassword$ = BLANK$
        Do
            nStatus% = cnReInitialize(TmpLogin)
        Loop While nStatus% = FAIL% And Len(TmpLogin.sPassword$) > 0

        If nStatus% = FAIL% Then
            server(nIndex%).PrivateData.nStatus% = INACTIVE%
            Exit Function
        End If

    End If

    cnInitializeServer% = SUCCEED%

End Function

Function cnNoReturn (nConnectionNumber As Integer, sCmd As String) As Integer

'*Purpose: Execute a SQL statement with no results returned.

'Declarations

    Dim nStatus As Integer

'End Declarations

'Execute Command
    nStatus = cnReturn(nConnectionNumber%, sCmd$)
    If nStatus = FAIL% Then
        Exit Function
    End If
    
' Note:  SQLFetch is not valid with a SQL command that
' does not return any results.
    
    cnNoReturn% = SUCCEED%

End Function

Function cnODBCError (nConnectionNumber As Integer) As Integer

'*Purpose: Generic error handler for ODBC errors

'Declarations

    Const SBufferLen = 256            ' Default String Buffer Length

    Dim nErrorCode As Integer
    Dim nReturnCode As Integer
    Dim nStatus As Integer
    Dim nOutlen As Integer
    Dim nErrorGroup As Integer
    Dim nServerIndex As Integer

    Dim lNative As Long

    Dim sMessage As String
    Dim sSBuffer1 As String * 16
    Dim sSBuffer3 As String * SBufferLen

'End Declarations

    Select Case connection(nConnectionNumber%).nReturnCode

    Case SQL_INVALID_HANDLE
        nStatus% = cnErrorHandle((connection(nConnectionNumber%).nReturnCode), ODBC_GROUP%, (connection(nConnectionNumber%).sActivityDescription), "Invalid Handle.")
        cnODBCError% = EXIT_STATUS%
        Exit Function

    Case SQL_NO_DATA_FOUND
        nStatus% = cnErrorHandle((connection(nConnectionNumber%).nReturnCode), ODBC_GROUP%, (connection(nConnectionNumber%).sActivityDescription), "No data found.")
        cnODBCError% = EXIT_STATUS%
        Exit Function

    Case SQL_ERROR, SQL_SUCCESS_WITH_INFO

    Case Else
        nStatus% = cnErrorHandle((connection(nConnectionNumber%).nReturnCode), ODBC_GROUP%, (connection(nConnectionNumber%).sActivityDescription), "Unknown ODBC Error.")
        cnODBCError% = EXIT_STATUS%
        Exit Function

    End Select

  ' Print an error message for the given connection handle
  ' and statement handle
    sSBuffer1$ = String$(16, 0)
    sSBuffer3$ = String$(SBufferLen, 0)
    
    nStatus% = cnGetServerIndex(connection(nConnectionNumber%).sServer, nServerIndex%)
    
    Select Case server(nServerIndex%).PrivateData.sServerType$
    Case SQL_SERVER$
        nErrorGroup% = SYBASE_GROUP%
    Case ALLBASE$
        nErrorGroup% = ALLBASE_GROUP%
    Case Else
    End Select
    
    nReturnCode% = SQLError(connection(nConnectionNumber).henv, connection(nConnectionNumber).hdbc, connection(nConnectionNumber).hstmt, sSBuffer1$, lNative&, sSBuffer3$, SBufferLen, nOutlen%)
  
    Do While (nReturnCode% = SQL_SUCCESS% Or nReturnCode% = SQL_SUCCESS_WITH_INFO%)
        If nOutlen% = 0 Then
            sMessage$ = "Error -- No error information available"
        Else
            sMessage$ = Left$(sSBuffer3$, nOutlen%)
        End If
        
        nStatus% = cnErrorHandle(lNative&, nErrorGroup%, (connection(nConnectionNumber%).sActivityDescription), sMessage$)

        nReturnCode% = SQLError(connection(nConnectionNumber).henv, connection(nConnectionNumber).hdbc, connection(nConnectionNumber).hstmt, sSBuffer1$, lNative&, sSBuffer3$, SBufferLen, nOutlen%)
    Loop

    cnODBCError% = nStatus%
    
End Function

Function cnOpenConnection (nConnection As Integer, sActivityDescription As String) As Integer

'*Purpose: If only one server connected, allow to get a connection assuming want existing server.

'Declarations

    Dim nStatus As Integer
    Dim nCount As Integer

'End Declarations
    
    On Error GoTo nOpenConnection_Error

    nCount% = UBound(server)

    On Error GoTo 0

    If nCount% > 0 Then
        nStatus% = cnErrorHandle(MULTI_SERVER_ERROR%, SERVER_OBJECT_GROUP%, "cnOpenConnection", "Multiple Servers Connected, Cannot Default connection: Contact Programmer" & NL$ & "Contact Programmer")
        Exit Function
    End If
    
    nStatus% = cnOpenConnectionMulti(server(0).Login.sServer$, nConnection%, sActivityDescription$)
    If nStatus% = FAIL% Then
        Exit Function
    End If

    cnOpenConnection% = SUCCEED%

nOpenConnection_Exit:

    Exit Function

nOpenConnection_Error:

    nStatus% = cnErrorHandle(SERVER_NOT_FOUND%, SERVER_OBJECT_GROUP%, "cnOpenConnection", "No Servers Exist for Default Connection" & NL$ & "Contact Programmer")
    
    Resume nOpenConnection_Exit

End Function

Function cnOpenConnectionMulti (sThisServer As String, nThisConnection As Integer, sActivityDescription As String) As Integer

'Purpose: Return an id for an available connection (if one does not exist, create it), may specify server (multi-server environment).

'Declarations

    Dim nStatus As Integer

    Dim TmpLogin As LoginType

'End Declarations

'Check server status
'If no server specified, get a server (login)

    TmpLogin.sServer$ = sThisServer$
    
    nStatus% = cnOpenServer(TmpLogin)
    
    If nStatus% = FAIL% Then
        Exit Function
    End If
            
    sThisServer$ = TmpLogin.sServer$

'********************************
'Have a server
'Now get a conneciton

    nStatus% = cnFindInactiveConnection(sThisServer$, nThisConnection%)
    If nStatus% = FAIL% Then

'No Available connection, make one
        nStatus% = cnCreateNewConnection(sThisServer$, nThisConnection%)
        If nStatus% = FAIL% Then
            Exit Function
        End If
    End If

'Available: check if active or inactive status
'Inactive connection - reconnect
    If connection(nThisConnection%).nStatus% = INACTIVE% Then
        nStatus% = cnActivateConnection(nThisConnection%)
        If nStatus% = FAIL% Then
            Exit Function
        End If
    End If

'Stamp what doing with the connection
'Description for error usage
    connection(nThisConnection%).sActivityDescription$ = sActivityDescription$

'Check it out to user
    connection(nThisConnection%).nUse% = ACTIVE%

    cnOpenConnectionMulti% = SUCCEED%

End Function

Private Function cnOpenNewServer (ThisLogin As LoginType) As Integer

'*Purpose: Open as server for processing.

'Declarations

    Dim nStatus As Integer
    Dim nMaxServers As Integer

'End Declarations

'Re Dimension server array for new server
    On Error GoTo nOpenNewServer_Error
    
    nMaxServers% = UBound(server) + 1
    
    On Error GoTo 0
    
    ReDim Preserve server(nMaxServers%) As ServerType

    server(nMaxServers%).Login = ThisLogin

    nStatus% = cnReadServerODBCData(ThisLogin.sServer$, nMaxServers%)

    If nStatus% = FAIL% Then
        Exit Function
    End If

'Allocate environment
    nStatus% = cnAllocateEnvironment(server(nMaxServers%))
    
    If nStatus% = FAIL% Then
        Exit Function
    End If

    server(nMaxServers%).PrivateData.nStatus% = ACTIVE%

    ThisLogin = server(nMaxServers%).Login

    cnOpenNewServer% = SUCCEED%

    Exit Function

nOpenNewServer_Error:

'Trap for when array not dimensioned - UBOUND fails

    nMaxServers% = 0
    
    Resume Next

End Function

Function cnOpenServer (TmpLogin As LoginType) As Integer

'*Pupose: Connect to server

'Declarations

    Dim nStatus As Integer

    Dim sServerType As String

'End Declarations


'No server specified
    
    If TmpLogin.sServer$ = BLANK$ Then
' Get login to collect info
        nStatus% = cnShowLogin(TmpLogin)
        If nStatus% = FAIL% Then
            Exit Function
        End If
    End If

'****************************
'Have a server name

' check if open
    
    nStatus% = cnCheckOpenServer(TmpLogin.sServer$)
    
'If not open
    
    If nStatus% = FAIL% Then
'No login or password information
        If TmpLogin.sLogin$ = BLANK$ Or TmpLogin.sPassword$ = BLANK$ Then
            nStatus% = cnGetServerType(TmpLogin.sServer$, sServerType$)
'If allbase, will get all data when open new server
            If UCase(sServerType$) <> UCase(ALLBASE$) Then
                nStatus% = cnShowLogin(TmpLogin)
                If nStatus% = FAIL% Then
                    Exit Function
                End If
            End If
        End If

        nStatus% = cnOpenNewServer(TmpLogin)
    Else
'Already open, get current data
        nStatus% = cnGetLoginData(TmpLogin)
    End If
    
    cnOpenServer% = SUCCEED%

End Function

Private Function cnParseAllBaseLastUser (sLoginString As String, AllBaseLogin As AllBaseLoginType) As Integer

'*Purpose: Break up the login string into components.

'Declarations

    Dim nStart As Integer
    Dim nEnd As Integer

'End Declarations

    nStart% = 1

'find operating system end delimeter

    nEnd% = InStr(nStart%, sLoginString$, "/")

    AllBaseLogin.sSystem$ = Mid(sLoginString$, nStart%, nEnd% - nStart%)

'find server

    nStart% = nEnd% + 1

    nEnd% = InStr(nStart%, sLoginString$, ":")

    AllBaseLogin.sServer$ = Mid(sLoginString$, nStart%, nEnd% - nStart%)

'find path

    nStart% = nEnd% + 1

    nEnd% = InStr(nStart%, sLoginString$, "#")

    AllBaseLogin.sPath$ = Mid(sLoginString$, nStart%, nEnd% - nStart%)

'find login

    nStart% = nEnd% + 1

    nEnd% = InStr(nStart%, sLoginString$, ":")

    AllBaseLogin.sLogin$ = Mid(sLoginString$, nStart%, nEnd% - nStart%)

'get password
    
    nStart% = nEnd% + 1
    
    AllBaseLogin.sPassword$ = Mid(sLoginString$, nStart%)
    
    cnParseAllBaseLastUser% = SUCCEED%

End Function

Function cnReadServerODBCData (sServer As String, nIndex As Integer) As Integer

'*Purpose: Read Server info from the ODBC.ini file

'Declarations

    Const FileName = "ODBC.INI"

    Dim AllBaseLogin As AllBaseLoginType
    
    Dim nStatus As Integer

    Dim sTemp As String * 255
    Dim sSection As String

'End Declarations

'Get Server type

    sSection$ = "ODBC Data Sources"

    nStatus% = GetPrivateProfileString(sSection$, sServer$, "n/a", sTemp$, Len(sTemp$), FileName$)

    server(nIndex%).PrivateData.sServerType = csRStrip$(sTemp$)

    If Len(sTemp$) = 0 Then
        nStatus% = cnErrorHandle(SERVER_NOT_IN_ODBC_INI%, SERVER_OBJECT_GROUP%, "cnReadServerODBCData", "No Server Type Found" & NL$ & "Contact Programmer")
    End If

'If ALLBASE get last user

    
    Select Case UCase(server(nIndex%).PrivateData.sServerType)
    Case UCase(ALLBASE$)
    
        sSection$ = sServer$

        nStatus% = GetPrivateProfileString(sSection$, "LastUser", "", sTemp$, Len(sTemp$), FileName$)

        server(nIndex%).PrivateData.sLastUser$ = csRStrip$(sTemp$)

        If Len(sTemp$) = 0 Then
            nStatus% = cnErrorHandle(LAST_LOGIN_ODBC_INI_ERROR%, SERVER_OBJECT_GROUP%, "cnReadServerODBCData", "No ALLBASE Login Data Found" & NL$ & "Contact Programmer")
            Exit Function
        End If

'Parse login string to get login and password

        nStatus% = cnParseAllBaseLastUser(server(nIndex%).PrivateData.sLastUser$, AllBaseLogin)
        
        If nStatus% = SUCCEED% Then
            server(nIndex%).Login.sLogin$ = AllBaseLogin.sLogin$
            server(nIndex%).Login.sPassword$ = AllBaseLogin.sPassword$
        End If

    Case SQL_SERVER$
'If Sybase don't get anything
    End Select

    cnReadServerODBCData% = SUCCEED%

End Function

Private Function cnReInitialize (TmpLogin As LoginType) As Integer

'*Purpose: Relogin after login failure (bad password).

'Declarations

    Dim nStatus As Integer
    Dim nIndex As Integer
    Dim nConnection As Integer

'End Declarations

' Get login to collect info
    nStatus% = cnShowLogin(TmpLogin)
    
    If nStatus% = FAIL% Then
        Exit Function
    End If

'Save new login info to server
    
    nStatus% = cnGetServerIndex(TmpLogin.sServer$, nIndex%)
    If nStatus% = FAIL% Then
        Exit Function
    End If

    server(nIndex%).Login = TmpLogin

'Test connection login
    
    nStatus% = cnOpenConnectionMulti(TmpLogin.sServer$, nConnection%, "Re-Login Attempt")
    
    If nStatus% = SUCCEED% Then
        server(nIndex%).PrivateData.nStatus% = ACTIVE%
        nStatus% = cnGetDefaultIsolation(nConnection%)
        nStatus% = cnReturnConnection(nConnection%)
    Else
        Exit Function
    End If

    cnReInitialize% = SUCCEED%

End Function

Private Function cnRemoveAllConnections (sThisServer As String) As Integer

'*Purpose: Close all open connections to the database

'Declarations

    Dim x As Integer
    Dim nRemoveFlag As Integer
    Dim nLast As Integer
    
    Dim Status As Integer

'End Declarations

'If no connections then nothing to do.
    
    On Error GoTo nRemoveAllConnections_Error
    
    nLast% = UBound(connection)

    On Error GoTo 0

    For x% = 0 To nLast%

        nRemoveFlag% = False
        
        Select Case sThisServer$
        Case Is = connection(x%).sServer$
            nRemoveFlag% = True
        Case Is = sAll_SERVERS$
            nRemoveFlag% = True
        End Select
            
        If nRemoveFlag% Then
            If connection(x%).nStatus = ACTIVE% Then
                Status% = cnRemoveConnection(x%)
                If Status% = FAIL% Then
                    Exit Function
                End If
            End If
        End If
    Next x%

    cnRemoveAllConnections% = SUCCEED%

nRemoveAllConnections_Exit:

    Exit Function

nRemoveAllConnections_Error:

    Resume nRemoveAllConnections_Exit

End Function

Function cnRemoveConnection (nConnectionNumber As Integer) As Integer

'*Purpose: Remove a connection

'Declarations

    Dim Status As Integer

'End Declarations

'Check that have active connection
    If connection(nConnectionNumber%).nStatus = INACTIVE% Then
        Exit Function
    End If
'
'   Drop the connection.  The SQL_DROP feature should deallocate the memory associated with the statement handle.
'
    connection(nConnectionNumber%).nReturnCode = SQLFreeStmt(connection(nConnectionNumber%).hstmt, SQL_DROP)
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
        Status% = cnODBCError(nConnectionNumber%)
      
        Select Case Status
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select

    End If

'
'   Closes the connection associated with a specific connection handle.
'
    connection(nConnectionNumber%).nReturnCode = SQLDisconnect(connection(nConnectionNumber%).hdbc)
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
        Status% = cnODBCError((nConnectionNumber%))
      
        Select Case Status
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select

    End If
    
'
'  Releases the connection handle and frees all memory associated with the handle.
'
    connection(nConnectionNumber%).nReturnCode = SQLFreeConnect(connection(nConnectionNumber%).hdbc)
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
        Status% = cnODBCError(nConnectionNumber%)
      
        Select Case Status
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select

    End If

    connection(nConnectionNumber%).nStatus = INACTIVE%
    connection(nConnectionNumber%).nUse = INACTIVE%
    
    
    cnRemoveConnection% = SUCCEED%

End Function

Function cnReturn (nConnectionNumber As Integer, sCmd As String) As Integer

'*Purpose: Process a SQL statement and set up for procedure to loop the data.

'Declarations

    Dim nStatus As Integer
    Dim nStatus2 As Integer

'End Declarations

'   Clear out Connection

    On Error GoTo nReturn_Exit
'
    nStatus% = SQLFreeStmt(connection(nConnectionNumber%).hstmt, SQL_CLOSE)

    On Error GoTo 0

    connection(nConnectionNumber).nReturnCode = SQLExecDirect(connection(nConnectionNumber%).hstmt, sCmd$, Len(sCmd$))
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS Then
        nStatus% = cnODBCError(nConnectionNumber%)
        If connection(nConnectionNumber%).nReturnCode = SQL_ERROR% Then
            nStatus2% = cnErrorHandle(SQL_SYNTAX_ERROR%, SERVER_OBJECT_GROUP%, "cnReturn", sCmd$)
        End If
      
        Select Case nStatus%
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select
    End If
    
    cnReturn% = SUCCEED%

nReturn_Exit:

End Function

Function cnReturnConnection (nConnectionNumber As Integer) As Integer

'*Purpose: Free statement handle and mark connection as not in use.

'Declarations

    Dim nStatus As Integer

    Dim nIndex As Integer

'End Declarations

    On Error GoTo nReturnConnection_Error

    nStatus% = SQLFreeStmt(connection(nConnectionNumber%).hstmt, SQL_CLOSE)

    On Error GoTo 0

'Reset USE flag for being available
    
    connection(nConnectionNumber%).nUse% = INACTIVE%

'If over max allowed then kill the connection and set STATUS flag inactive.
    
    Select Case nConnectionNumber%
    Case Is >= MAX_OPEN_CONNECTIONS%
        nStatus% = cnRemoveConnection(nConnectionNumber%)
        If nStatus% = SUCCEED% Then
            connection(nConnectionNumber%).nStatus% = INACTIVE%
        End If
    Case Else
'NOTE: hmw 12/03/93 It is necessary to reset the connection to the default of autocommit
' If begin has been called, to turn autocommit off, there are problems if no explicit commits are made
' this is possible if one use of the connection does a begin/rollback/commit and the next open_connection does not (since the first connection is not killed between uses)
        If connection(nConnectionNumber%).lAutoCommit& = AUTOCOMMIT_OFF& Then
            connection(nConnectionNumber%).nReturnCode = SQLSetConnectOption(connection(nConnectionNumber%).hdbc, SQL_AUTOCOMMIT, AUTOCOMMIT_ON&)
            If connection(nConnectionNumber%).nReturnCode% <> SQL_SUCCESS% Then
                nStatus% = cnODBCError(nConnectionNumber%)
                Select Case nStatus%
                Case EXIT_STATUS%, FAIL%
                    Exit Function
                End Select
            End If
            connection(nConnectionNumber%).lAutoCommit& = AUTOCOMMIT_ON&
        End If
    
        nStatus% = cnGetServerIndex(connection(nConnectionNumber%).sServer$, nIndex%)
        
'Reset isolation level to default
        If connection(nConnectionNumber%).lIsolationLevel& <> server(nIndex%).PrivateData.lIsolationLevel& Then
            connection(nConnectionNumber%).nReturnCode = SQLSetConnectOption(connection(nConnectionNumber%).hdbc, SQL_TXN_ISOLATION, server(nIndex%).PrivateData.lIsolationLevel&)
            If connection(nConnectionNumber%).nReturnCode% <> SQL_SUCCESS% Then
            nStatus% = cnODBCError(nConnectionNumber%)
                Select Case nStatus%
                Case EXIT_STATUS%, FAIL%
                    Exit Function
                End Select
            End If
        End If
    End Select

    cnReturnConnection% = SUCCEED%

nReturnConnection_Exit:

    Exit Function

nReturnConnection_Error:

    Resume nReturnConnection_Exit

End Function

Function cnRollbackTrans (nConnectionNumber As Integer) As Integer

'*Purpose: Committ a SQL transaction

'Declarations

    Dim nStatus As Integer
    Dim nIndex As Integer

    Dim sCmd As String

'End Declarations
 
    nStatus% = cnGetServerIndex(connection(nConnectionNumber%).sServer$, nIndex%)
    If nStatus% = FAIL% Then
        Exit Function
    End If

'    Select Case UCase(server(nIndex%).PrivateData.sServerType$)
'    Case UCase(SQL_SERVER$)
'        sCmd$ = "rollback tran"
'        nStatus% = cnNoReturn(nConnectionNumber%, sCmd$)
'    Case UCase(ALLBASE$)
        connection(nConnectionNumber%).nReturnCode% = SQLTransact(connection(nConnectionNumber%).henv, connection(nConnectionNumber%).hdbc, SQL_ROLLBACK)
        If connection(nConnectionNumber%).nReturnCode% <> SQL_SUCCESS% Then
            nStatus% = cnODBCError(nConnectionNumber%)
        End If
'    End Select
    
    Select Case nStatus%
    Case EXIT_STATUS%, FAIL%
        Exit Function
    End Select
    
    cnRollbackTrans% = SUCCEED%

End Function

Private Function cnSetDatabase (Data_Base As String, nConnection As Integer) As Integer

'*Purpose: set connection to point to a database

'Declarations

    Dim nStatus As Integer

    Dim sCmd As String

'End Declarations

    If Data_Base$ = BLANK$ Then
        Exit Function
    End If

    sCmd$ = "use " + Data_Base$

    nStatus% = cnNoReturn(nConnection%, sCmd$)
    
    Select Case nStatus%
    Case SUCCEED%
    Case Else
        Exit Function
    End Select
    
    cnSetDatabase% = SUCCEED%

End Function

Function cnSetInitials (sTmpInitial As String) As Integer

'*Purpose: set the server initials

    sInitial$ = sTmpInitial$

    cnSetInitials% = SUCCEED%

End Function

Function cnSetPassLogin (ThisServer As PassLoginType) As Integer

    PassLogin = ThisServer

    cnSetPassLogin% = SUCCEED%

End Function

Private Function cnShowLogin (ThisLogin As LoginType) As Integer

'*Purpose: Set pass data and show login form

'Declarations

    Dim nStatus As Integer

'End Declarations

'Load server parameters to pass structure

    PassLogin.Login = ThisLogin

'NOTE: Comment out one or the other of the following lines
'ILX does not use the login form, and the module may not be available
'If using SYBASE, then comment out the second
'IF using ALLBASE (ILX) then comment out the first.

    frmLogin.Show MODAL%
'    nStatus% = cnErrorHandle(LOGIN_FORM_REQUEST%, SERVER_OBJECT_GROUP%, "server object", "Insufficient login data so requesting login form, invalid for sevrer type")

    ThisLogin = PassLogin.Login

    If ThisLogin.sLogin = BLANK$ Then
        Exit Function
    End If

    cnShowLogin% = SUCCEED%

End Function

Function cnShutDown () As Integer

'*Purpose: End ODBC session.

'Declarations

    Dim nStatus As Integer

'End Declarations

'Close All Connections
    nStatus% = cnRemoveAllConnections(sAll_SERVERS$)

    If nStatus% = FAIL Then
        Exit Function
    End If

'DeAllocate Environment
    nStatus% = cnDeAllocateEnvironment(sAll_SERVERS$)
      
    If nStatus% = FAIL% Then
       Exit Function
    End If

    ReDim connection(0)

    ReDim server(0)

    cnShutDown% = SUCCEED%
    
End Function

Function csGetData (nConnectionNumber As Integer, nColumn As Integer, nReturnStatus As Integer) As String

'*Purpose: Return data via csGetLargeData - default data size to std small size.

'Declarations
    
    Const nSmallSize% = 256

'End Declarations

    csGetData$ = csGetLargeData(nConnectionNumber%, nColumn%, nSmallSize%, nReturnStatus%)
    
End Function

Function csGetLargeData (nConnectionNumber As Integer, nColumn As Integer, nDataSize As Integer, nReturnStatus As Integer) As String

'*Purpose: Return data via an ODBC call

'Declarations
    
    Dim lSize As Long         'return length of buffer (hold)

    Dim sHold As String       'data buffer
    Dim nStatus As Integer
    
'End Declarations

    sHold$ = Space$(nDataSize%)

    nReturnStatus% = FAIL%

    connection(nConnectionNumber%).nReturnCode = SQLGetData(connection(nConnectionNumber%).hstmt, nColumn%, SQL_C_CHAR, sHold$, Len(sHold$), lSize&)
    If connection(nConnectionNumber%).nReturnCode <> SQL_SUCCESS% Then
        nStatus% = cnODBCError(nConnectionNumber%)
      
        Select Case nStatus
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select

    End If

    If lSize& = SQL_NULL_DATA Then
       csGetLargeData$ = BLANK$
    Else
       csGetLargeData$ = RTrim$(Left$(sHold$, lSize&))
    End If
    
    nReturnStatus% = SUCCEED%

End Function

Function csLastReturn (nConnectionNumber As Integer, sCmd As String, nReturnStatus As Integer) As String

'*Purpose: Return the last result (used with stored procedures that may have multiple returns.

'Declarations

    Const nFirstColumn% = 1
    
    Dim nResultCode As Integer
    Dim nFetchCode As Integer
    Dim nDataCode As Integer
    Dim nStatus As Integer

    Dim lCount As Long
    
'End Declarations

    nReturnStatus% = FAIL%

'Execute Command
    nStatus% = cnReturn(nConnectionNumber%, sCmd$)
    If nStatus% = FAIL% Then
        Exit Function
    End If
    
'Loop all possible Statements in buffer
    Do While cnGetResults(nConnectionNumber%) = SQL_SUCCESS%
        
'Loop all data in data set
        Do While cnGetRow(nConnectionNumber%) = SQL_SUCCESS%

'Get data from data set
            csLastReturn$ = csGetData(nConnectionNumber%, nFirstColumn%, nDataCode%)
            If nDataCode% = FAIL% Then
                Exit Function
            End If
            
'Get next row

        Loop

'Check successful completion
        If connection(nConnectionNumber%).nReturnCode <> SQL_NO_DATA_FOUND% Then
            nStatus% = cnODBCError(nConnectionNumber%)
      
            Select Case nStatus%
            Case EXIT_STATUS%, FAIL%
                Exit Function
            End Select

        End If

'Get next SQL result set
        connection(nConnectionNumber%).nReturnCode = SQLMoreResults%(connection(nConnectionNumber%).hstmt)

    Loop
    
'Check successful completion
    If connection(nConnectionNumber%).nReturnCode <> SQL_NO_DATA_FOUND% Then
        nStatus% = cnODBCError(nConnectionNumber%)
      
        Select Case nStatus%
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select

    End If

    connection(nConnectionNumber%).nReturnCode = SQL_SUCCESS%
    
    nReturnStatus% = SUCCEED%

End Function

Function csOneReturn (nConnectionNumber As Integer, sCmd As String, nReturnStatus As Integer) As String

'*Purpose: Return one value from a SQL call.

'Declarations

    Const nFirstColumn% = 1

    Dim nDataCode As Integer
    Dim nStatus As Integer

'End Declarations

    nReturnStatus% = FAIL%

'Execute Command

    nStatus = cnReturn(nConnectionNumber%, sCmd$)
    If nStatus = FAIL% Then
        Exit Function
    End If
    
'Loop all data in data set
    Do While cnGetRow(nConnectionNumber%) = SQL_SUCCESS%
'Get data from data set
        csOneReturn$ = csGetData(nConnectionNumber%, nFirstColumn%, nStatus%)

        If nStatus = FAIL% Then
            Exit Function
        End If

    Loop

'Check successful completion
    If connection(nConnectionNumber%).nReturnCode <> SQL_NO_DATA_FOUND% Then
        nStatus% = cnODBCError(nConnectionNumber%)
      
        Select Case nStatus
        Case EXIT_STATUS%, FAIL%
            Exit Function
        End Select
    End If
    
    connection(nConnectionNumber%).nReturnCode% = SQL_SUCCESS%
    
    nReturnStatus% = SUCCEED%

End Function

